function [W,B,C,Py,H,f,iter,lossX,lossY]=t3pcr(X,m1,m2,m3,Y,q,r1,r2,r3,alfa,start,conv,W,B,C);
% [W,B,C,Py,H,f,iter,lossX,lossY]=t3pcr(X,m1,m2,m3,Y,q,r1,r2,r3,alfa,start,conv,W,B,C);
%
% Procedure for Principal Covariates Regression, with 3-way TUCKALS model for X
%
% uses m-files: permnew.m, ssq.m, ed.m,  
%
% Written by Henk A.L. Kiers, University of Groningen (Last update: September 1, 1999)
%
% Minimizes f(W,B,C,H,By) = alfa* || X - (X W) H (C'xB')||^2 + (1-alfa)* || Y - X W Py ||^2
% where x means Kronecker product, and X refers to the three-way array
%
% Input : 
%    X = (m1 x m2 x m3) arranged in supermatrix of order m1 x m2*m3 with first column index running fastest, last slowest
%    Y = (m1 x q) matrix with dependent variables
%    m1,m2,m3: order of X
%    q: number of columns in Y
%    r1,r2,r3: dimensionality of candecomp/parafac
%    alfa = pcr parameter (between 0 and 1)
%    conv = convergence criterion
%    start 
%       0 = A=XW and B, C from svds (default)
%       1 = Random (orthonormalized) start
%       2 = Present values of W, B, and C
%
% Output: 
%    W,B,C,Py (see loss function)
%    f = loss function value
%    iter = number of iterations 
%    lossX = || X - (X W) H (C'xB')||^2
%    lossY = || Y - X W Py ||^2
%
% Further comments:
% Note that the matrix A is forced to be in the column space of X
% An alternative procedure would be to leave A completely free!
% At present, the loss for X has as lower bound the tucker3 loss.
% The loss for Y has as lower bound the Reduced Rank regression loss, where a 
% PCA with respect to X is performed on Y.
%   ==========================================================================
%   Copyright 2005 Biosystems Data Analysis Group ; Universiteit van Amsterdam
%   ==========================================================================


prntlag=1;  % frequency of screen update

[nn,mprod]=size(X);
if nn~=m1 | mprod~=m2*m3, 
   fprintf(' Specified size of array does not correspond to actual size\n');
   keyboard
end

ssx=ssq(X);
ssy=ssq(Y);

if start==0
   % W and auxiliary matrix A=XW (note A columnw. orthon.);
   if m1>m2*m3
      [P,Dd,Q]=svd(X,0); 
      W=Q(:,1:r1)/Dd(1:r1,1:r1); 
   else
      [P,Dd]=ed(X*X');
   end
   A=P(:,1:r1);    
   
   % Py
   Py=A\Y;   
   
   % B
   Z=permnew(X,m1,m2,m3);
   if m2>m1*m3  
      [P,Dd,Q]=svd(Z,0); 
   else
      [P,Dd]=ed(Z*Z');
   end
   B=P(:,1:r2);
   
   % C
   Z=permnew(Z,m2,m3,m1);
   if m3>m1*m2
      [P,Dd,Q]=svd(Z,0); 
   else
      [P,Dd]=ed(Z*Z');
   end
   C=P(:,1:r3);
end

if start==1
   disp(' random  O R T H O N O R M A L I Z E D  starts')
   W=rand(m2*m3,r1)-.5;
   [A,R]=qr(X*W,0);
   W=W/R;
   Py=A\Y;
   if m2>=r2,B=orth(rand(m2,r2)-.5); else B=orth(rand(r2,r2)-.5);B=B(1:m2,:); end
   if m3>=r3,C=orth(rand(m3,r3)-.5); else C=orth(rand(r3,r3)-.5);C=C(1:m3,:); end
end

if start==2
   A=X*W;
   Py=A\Y;
end

% Update H 
Z1=permnew(X,m1,m2,m3);
Z1=permnew(B'*Z1,r2,m3,m1);
Z1=permnew(C'*Z1,r3,m1,r2);  % gives m1 x r2*r3 array
H=A\Z1;

H1=permnew(H,r1,r2,r3);
H1=permnew(B*H1,m2,r3,r1);
H1=permnew(C*H1,m3,r1,m2);  % gives r1 x m2*m3 array  H*C'xB'
lossX=ssq(X-A*H1);
lossY=ssq(Y-A*Py);
f=alfa*lossX+(1-alfa)*lossY;
fprintf('function value at Start is %12.8f \n',f)

fold=f+2*conv*f;
iter=0;

while fold-f>conv*f | iter<2
   
   fold=f;  
   iter=iter+1;
   
   % Update A
   
   % min alfa*|| X - XW H C'xB' ||^2 + (1-alfa)* || Y - XW Py ||^2
   % minimize ||(alfa^.5*X|(1-alfa)^.5*Y) - X W (alfa^.5*F|(1-alfa)^.5)*Py) ||
   % where F = H C'xB'
   
   FF=alfa*H*H'+(1-alfa)*Py*Py';
   Z1=permnew(X,m1,m2,m3);
   Z1=permnew(B'*Z1,r2,m3,m1);
   Z1=permnew(C'*Z1,r3,m1,r2);  % gives m1 x r2*r3 array
   ZF=Z1*H';
   if m2*m3>=m1
      A=(alfa*ZF+(1-alfa)*Y*Py')/FF;
   else
      W=X\(alfa*ZF+(1-alfa)*Y*Py')/FF;
      A=X*W;
   end
   
   % Update Py
   Py=A\Y;
   
   % Update B
   Z=permnew(X,m1,m2,m3);
   Z1=permnew(Z,m2,m3,m1);
   Z1=permnew(C'*Z1,r3,m1,m2);
   Z1=permnew(A'*Z1,r1,m2,r3);  % gives m2 x r1*r3 array
   H=permnew(H,r1,r2,r3);
   Z1Ht=Z1*H';
   [P,Ds,Q]=svd(Z1Ht,0);
   B=P*Q';
   
   %f=alfa*ssq(Z-B*F)+(1-alfa)*ssq(Y-A*Py);
   %  fprintf(' f after updating B is %12.8f \n',f);
   
   % Update C
   Z=permnew(Z,m2,m3,m1);
   Z1=permnew(Z,m3,m1,m2);
   Z1=permnew(A'*Z1,r1,m2,m3);
   Z1=permnew(B'*Z1,r2,m3,r1);  % gives m3 x r1*r2 array
   H=permnew(H,r2,r3,r1);
   [P,Ds,Q]=svd(Z1*H',0);
   C=P*Q';
   
   %f=alfa*ssq(Z-C*F)+(1-alfa)*ssq(Y-A*Py);
   %fprintf(' f after updating C is %12.8f \n',f);
   
   % Update H
   Z1=permnew(X,m1,m2,m3);
   Z1=permnew(B'*Z1,r2,m3,m1);
   Z1=permnew(C'*Z1,r3,m1,r2);  % gives m1 x r2*r3 array
   H=A\Z1;
   
   % Evaluate
   lossX=ssx-trace((A'*A)*(H*H'));
   lossY=ssq(Y-A*Py);
   f=alfa*lossX+(1-alfa)*lossY;
   if rem(iter,prntlag)==0
      fprintf('f = %12.8f after %g iters; lossX= %10.6f lossY=%10.6f \n',f,iter,lossX,lossY)
   end
   
end

% after convergence, compute W
if m2*m3>=m1
   W=pinv(X)*A;
end
fprintf('function value is %12.8f after %g iterations \n',f,iter)